import os
from typing import Any

from attr import define

from detectpii.resolver.model import Resolver


@define(kw_only=True)
class EnvironmentResolver(Resolver):
    """Resolves credentials from environment variables"""

    name: str = "EnvironmentResolver"

    def resolve(self, credential: Any) -> Any:
        return os.environ.get(credential)



from tabulate import tabulate

from detectpii.model import PiiColumn


def print_columns(pii_columns: list[PiiColumn]) -> None:
    rows = [
        [column.table, column.column, column.pii_type.name] for column in pii_columns
    ]

    print(tabulate(rows, headers=["Table", "Column", "Type"]))



from functools import cached_property
from typing import Sequence

from attr import define
from sqlalchemy import Engine, create_engine, text, Connection, Row, MappingResult
from trino.sqlalchemy import URL

from detectpii.model import Catalog, Column, Table


@define(kw_only=True)
class TrinoCatalog(Catalog):
    """A collection of tables in a Trino catalog and schema."""

    user: str
    password: str
    host: str
    port: int
    catalog: str
    schema: str

    @cached_property
    def engine(self) -> Engine:
        return create_engine(self.url, pool_timeout=10)

    def detect_tables(self) -> None:
        trino_engine = self.engine

        with trino_engine.connect() as conn:

            for row in self._table_names(conn=conn):
                (table_name,) = row

                self._add_table_to_catalog(
                    table_name=table_name,
                    conn=conn,
                )

    def sample(self, table: Table, percentage: int = 10, **kwargs) -> MappingResult:
        assert table in self.tables, "Provided table does not belong to this catalog"

        fully_qualified_table_name = f"{self.catalog}.{self.schema}.{table.name}"
        sql = text(
            f"""
            SELECT * 
            FROM {fully_qualified_table_name} 
            TABLESAMPLE BERNOULLI({percentage})
        """
        )

        with self.engine.connect() as conn:
            return conn.execute(
                statement=sql,
            ).mappings()

    @property
    def url(self) -> URL:
        return URL(
            host=self.resolver.resolve(self.host),
            user=self.resolver.resolve(self.user),
            port=self.resolver.resolve(self.port),
            password=self.resolver.resolve(self.password),
            catalog=self.resolver.resolve(self.catalog),
            schema=self.resolver.resolve(self.schema),
        )

    def _table_names(self, conn: Connection) -> Sequence[Row]:
        query = text("SHOW TABLES")
        return conn.execute(statement=query).fetchall()

    def _add_table_to_catalog(self, table_name: str, conn: Connection) -> None:
        query = text(f"DESCRIBE {table_name}")

        columns = [
            Column(name=column[0])
            for column in conn.execute(statement=query).fetchall()
        ]

        table = Table(
            name=table_name,
            schema=self.schema,
            columns=columns,
        )

        self.tables.append(table)



import itertools
from functools import cached_property

from attr import define
from sqlalchemy import Engine, create_engine, text, MappingResult

from detectpii.model import Catalog, Table, Column


@define(kw_only=True)
class SnowflakeCatalog(Catalog):
    """A collection of tables in a Snowflake database and schema."""

    user: str
    password: str
    account_identifier: str
    database: str
    schema: str

    @cached_property
    def engine(self) -> Engine:
        return create_engine(self.url)

    def detect_tables(self) -> None:
        with self.engine.connect() as conn:
            query = text(
                f"""
                SELECT table_name, column_name
                FROM information_schema.columns
                WHERE 1=1
                    AND LOWER(table_catalog) = LOWER('{self.database}')
                    AND LOWER(table_schema) = LOWER('{self.schema}')
                ORDER BY table_name, column_name
            """
            )

            rows = conn.execute(query)

            for table_name, table_rows in itertools.groupby(
                iterable=rows,
                key=lambda row: row[0],
            ):
                table = Table(
                    name=table_name,
                    schema=self.schema,
                    columns=[Column(name=table_row[-1]) for table_row in table_rows],
                )

                self.tables.append(table)

    def sample(
        self,
        table: Table,
        percentage: int = 10,
        *args,
        **kwargs,
    ) -> MappingResult:
        sql = text(
            f"""
            SELECT *
            FROM {table.name}
            TABLESAMPLE BERNOULLI(:percentage)
        """
        )

        with self.engine.connect() as conn:
            return conn.execute(
                statement=sql,
                parameters={
                    "percentage": percentage,
                },
            ).mappings()

    @property
    def url(self) -> str:
        return "snowflake://{user}:{password}@{account_identifier}/{database}/{schema}".format(
            user=self.resolver.resolve(self.user),
            password=self.resolver.resolve(self.password),
            account_identifier=self.resolver.resolve(self.account_identifier),
            database=self.resolver.resolve(self.database),
            schema=self.resolver.resolve(self.schema),
        )



import abc
import re
from typing import Optional

from attr import define

from detectpii.model import Column
from detectpii.pii_type import (
    Email,
    BirthDate,
    Gender,
    Nationality,
    ZipCode,
    UserName,
    SSN,
    PoBox,
    CreditCard,
    Phone,
    PiiType,
    Person,
    Address,
    Password,
)


@define(kw_only=True, frozen=True)
class Detector:

    @abc.abstractmethod
    def detect(
        self,
        column: Column,
        *args,
        **kwargs,
    ) -> Optional[PiiType]:
        raise NotImplementedError()


@define(kw_only=True)
class ColumnNameRegexDetector(Detector):
    """Detect PII columns by matching them against known patterns.

    This class has been borrowed from piicatcher and modified for use with this library.
    """

    regex: dict = {
        Person: re.compile(
            "^.*(firstname|fname|lastname|lname|"
            "fullname|maidenname|_name|"
            "nickname|name_suffix|name|person).*$",
            re.IGNORECASE,
        ),
        Email: re.compile("^.*(email|e-mail|mail).*$", re.IGNORECASE),
        BirthDate: re.compile(
            "^.*(date_of_birth|dateofbirth|dob|"
            "birthday|date_of_death|dateofdeath|birthdate).*$",
            re.IGNORECASE,
        ),
        Gender: re.compile("^.*(gender).*$", re.IGNORECASE),
        Nationality: re.compile("^.*(nationality).*$", re.IGNORECASE),
        Address: re.compile(
            "^.*(address|city|state|county|country|zone|borough).*$",
            re.IGNORECASE,
        ),
        ZipCode: re.compile(
            "^.*(zipcode|zip_code|postal|postal_code|zip).*$",
            re.IGNORECASE,
        ),
        UserName: re.compile("^.*user(id|name|).*$", re.IGNORECASE),
        Password: re.compile("^.*pass.*$", re.IGNORECASE),
        SSN: re.compile(
            "^.*(ssn|social_number|social_security|"
            "social_security_number|social_security_no).*$",
            re.IGNORECASE,
        ),
        PoBox: re.compile("^.*(po_box|pobox).*$", re.IGNORECASE),
        CreditCard: re.compile(
            "^.*(credit_card|cc_number|cc_num|creditcard|"
            "credit_card_num|creditcardnumber).*$",
            re.IGNORECASE,
        ),
        Phone: re.compile(
            "^.*(phone|phone_number|phone_no|phone_num|"
            "telephone|telephone_num|telephone_no).*$",
            re.IGNORECASE,
        ),
    }

    name: str = "ColumnNameRegexDetector"

    def detect(
        self,
        column: Column,
        *args,
        **kwargs,
    ) -> Optional[PiiType]:
        for pii_type, ex in self.regex.items():
            if ex.match(column.name) is not None:
                return pii_type()

        return None



import itertools
from functools import cached_property

from attr import define
from sqlalchemy import Engine, create_engine, text, URL, MappingResult

from detectpii.model import Catalog, Column, Table


@define(kw_only=True)
class PostgresCatalog(Catalog):
    """A collection of tables in a Postgres database and schema."""

    user: str
    password: str
    host: str
    port: int
    database: str
    schema: str

    @cached_property
    def engine(self) -> Engine:
        return create_engine(self.url)

    def detect_tables(self) -> None:
        query = text(
            f"""
            SELECT
                table_schema,
                table_name,
                column_name
            FROM
                information_schema.columns
            WHERE 1=1 
                  AND table_schema = :schema
            ORDER BY table_name, column_name;
        """
        )

        pg_engine = self.engine

        with pg_engine.connect() as conn:
            rows = conn.execute(
                statement=query,
                parameters={"schema": self.schema},
            )

            for table_name, table_rows in itertools.groupby(
                iterable=rows,
                key=lambda row: row[1],
            ):
                table = Table(
                    name=table_name,
                    schema=self.schema,
                    columns=[Column(name=table_row[-1]) for table_row in table_rows],
                )

                self.tables.append(table)

    def sample(
        self,
        table: Table,
        percentage: int = 10,
        *args,
        **kwargs,
    ) -> MappingResult:
        sql = text(
            f"""
            SELECT *
            FROM {table.name}
            TABLESAMPLE BERNOULLI(:percentage)
        """
        )

        with self.engine.connect() as conn:
            return conn.execute(
                statement=sql,
                parameters={
                    "percentage": percentage,
                },
            ).mappings()

    @property
    def url(self) -> URL:
        return URL(
            username=self.resolver.resolve(self.user),
            password=self.resolver.resolve(self.password),
            database=self.resolver.resolve(self.database),
            port=self.resolver.resolve(self.port),
            host=self.resolver.resolve(self.host),
            query={},
            drivername="postgresql+psycopg2",
        )



from typing import Any

from attr import define

from detectpii.resolver.model import Resolver


@define(kw_only=True)
class PlaintextResolver(Resolver):
    """Resolves credentials provided in plaintext."""

    name: str = "PlaintextResolver"

    def resolve(self, credential: Any) -> Any:
        return credential


@define(kw_only=True)
class NoneResolver(Resolver):
    """Placeholder resolver for creating a tagged union."""

    name: str = "NoneResolver"

    def resolve(self, credential: Any) -> Any:
        return None



from attr import define

from detectpii.detector import ColumnValueRegexDetector
from detectpii.model import Catalog, PiiColumn, Scanner, Column


@define(kw_only=True)
class DataScanner(Scanner):
    """Scan the table values for PII columns."""

    times: int = 1
    percentage: int = 10

    def scan(self, catalog: Catalog, **kwargs) -> list[PiiColumn]:
        pii_columns = []

        column_value_regex_detector = ColumnValueRegexDetector()

        # -- Repeat the process as many times as requested
        for _ in range(self.times):

            # -- One table at a time
            for table in catalog.tables:

                # -- One row at a time
                for row in catalog.sample(
                    table=table,
                    percentage=self.percentage,
                ):

                    # -- One column at a time
                    for column_name, value in row.items():
                        column = Column(name=column_name)

                        pii_type = column_value_regex_detector.detect(
                            column=column,
                            sample=str(value),
                        )

                        if pii_type:
                            pii_column = PiiColumn(
                                table=table.name,
                                column=column.name,
                                pii_type=pii_type,
                            )

                            pii_columns.append(pii_column)

        return pii_columns



from .trino_catalog import TrinoCatalog
from .postgres_catalog import PostgresCatalog
from .yugabyte_catalog import YugabyteCatalog
from .snowflake_catalog import SnowflakeCatalog

CatalogT = PostgresCatalog | TrinoCatalog | YugabyteCatalog | SnowflakeCatalog



from attr import define, Factory

from detectpii.catalog import CatalogT
from detectpii.model import PiiColumn
from detectpii.scanner import ScannerT


@define(kw_only=True)
class PiiDetectionPipeline:
    catalog: CatalogT
    scanners: list[ScannerT] = Factory(list)

    def scan(self) -> list[PiiColumn]:
        self.catalog.detect_tables()
        pii_columns = []

        for scanner in self.scanners:
            pii_columns.extend(scanner.scan(catalog=self.catalog))

        return list(set(pii_columns))
